@@ -24,6 +24,7 @@
 (require 'org-mem)
 (require 'magit-section)
 (require 'repeat)
+(require 'map)
 (eval-when-compile
   (require 'org)
   (require 'org-node)
@@ -327,6 +328,7 @@ Repeatable on the last key of a key sequence if
 Call the former if `org-node-context-follow-mode' is enabled,
 otherwise call the latter."
   (interactive)
+  (org-node-context--clean-stale-previews)
   (if org-node-context-follow-mode
       (org-node-context-toggle)
     (org-node-context-raise)))
@@ -424,7 +426,6 @@ that buffer."
         (insert (org-node-context--get-preview link))
         (insert "\n")))))
 
-(defvar org-node-context--snippet-link)
 (defun org-node-context--get-preview (link)
   "Get a preview snippet from around LINK in its containing file.
 
@@ -432,40 +433,28 @@ Actually, if a snippet was previously cached, return the cached version,
 else briefly visit the file, go to the link and call
 `org-node-context--extract-entry-at-point'."
   (let* ((entry (org-node-context--get-link-origin link))
-         (eid (or (org-mem-entry-id entry) (org-mem-entry-pseudo-id entry)))
+         (key (or (org-mem-entry-id entry) (org-mem-entry-pseudo-id entry)))
          (link-pos (org-mem-link-pos link))
-         ;; NOTE: `pos-diff' is not necessary in a simple implementation, but
-         ;; this level of granularity lets us avoid wiping all cached previews
+         ;; Tracking `pos-diff' lets us avoid wiping all cached previews
          ;; in a large file every time it is saved -- doing so would make the
          ;; cache useless, when you are working in a large file with links
          ;; between parts of itself.
-         ;;
-         ;; Instead, we just don't wipe anything, and trust in a sloppy rule of
-         ;; thumb: when the text between a link and its heading get edited,
-         ;; that will almost always result in a new unique `pos-diff'.
          (pos-diff (- link-pos (org-mem-entry-pos entry))))
-    (or (alist-get pos-diff (gethash eid org-node-context--previews))
-        (setf
-         (alist-get pos-diff (gethash eid org-node-context--previews))
-         (let (snippet)
-           (with-current-buffer (org-node--work-buffer-for (org-mem-file entry))
-             (goto-char link-pos)
-             (setq snippet (org-node-context--extract-entry-at-point)))
-           (with-current-buffer (org-mem-org-mode-scratch)
-             (erase-buffer)
-             (insert snippet)
-             (goto-char pos-diff)
-             (run-hooks 'org-node-context-postprocess-hook)
-             ;; Finally font-lock now that we are in a tiny buffer that
-             ;; contains only the snippet that needs to be font-locked, not the
-             ;; entire source file.
-             ;;
-             ;; It would be nice to do this before the postprocess hook instead
-             ;; of after, to offer the possibility to override colors
-             ;; or something, but the hook also could be used to add text,
-             ;; so we have to re-fontify in any case.
-             (font-lock-ensure)
-             (buffer-string)))))))
+    (with-memoization (alist-get pos-diff (gethash key org-node-context--previews))
+      (let (snippet)
+        (with-current-buffer (org-node--work-buffer-for (org-mem-file entry))
+          (goto-char link-pos)
+          (setq snippet (org-node-context--extract-entry-at-point)))
+        (with-current-buffer (org-mem-org-mode-scratch)
+          (erase-buffer)
+          (insert snippet)
+          (goto-char pos-diff)
+          (run-hooks 'org-node-context-postprocess-hook)
+          ;; Finally font-lock now that we are in a tiny buffer that
+          ;; contains only the snippet that needs to be font-locked, not the
+          ;; entire source file.
+          (font-lock-ensure)
+          (buffer-string))))))
 
 (defun org-node-context--extract-entry-at-point ()
   "Return whole entry at point as a string, including heading if any."
@@ -499,62 +488,48 @@ entry has no ID."
   "Whether to sync cached backlink previews to disk.
 The disk location is `org-node-data-dir'.
 
-This allows the context buffer created by \\[org-node-context-raise] to
-show up more instantly, even the first time it renders a given set of
-backlinks.
-
 Here\\='s how you test whether this is noticeable on your system:
-1. Open a file with many subtree nodes
-2. Collapse them all (i.e. S-TAB until you see only the headings)
-3. Have a context buffer open with `org-node-context-follow-mode'
-4. Hold the C-n or <down> key."
+1. Restart Emacs
+2. Open a file with many subtree nodes
+3. Collapse them all (i.e. S-TAB until you see only the headings)
+4. Have a context buffer open with `org-node-context-follow-mode'
+5. Hold the \\[next-line] key."
   :type 'boolean
   :package-version '(org-node . "2.0.0"))
 
 (defvar org-node-context--persist-timer (timer-create))
-(defvar org-node-context--last-tbl-state 0)
-(defvar org-node-context--did-init-persist nil)
 
 (defun org-node-context--maybe-init-persistence (&rest _)
   "Try to restore `org-node-context--previews' from disk.
 Then start occasionally syncing back to disk.
 No-op if user option `org-node-context-persist-on-disk' is nil."
-  (when org-node-context-persist-on-disk
-    (unless org-node-context--did-init-persist
-      (setq org-node-context--did-init-persist t)
-      (cancel-timer org-node-context--persist-timer)
-      (setq org-node-context--persist-timer
-            (run-with-idle-timer 50 t #'org-node-context--persist))
-      ;; Load from disk.
-      (when (file-readable-p (org-node-context--persist-file))
-        (with-temp-buffer
-          (insert-file-contents (org-node-context--persist-file))
-          (let ((data (read (current-buffer))))
-            (when (hash-table-p data)
-              (setq org-node-context--last-tbl-state (hash-table-count data))
-              (setq org-node-context--previews data))))))))
+  (when (and org-node-context-persist-on-disk
+             (hash-table-empty-p org-node-context--previews))
+    (cancel-timer org-node-context--persist-timer)
+    (setq org-node-context--persist-timer
+          (run-with-idle-timer 30 t #'org-node-context--persist))
+    (org-node-context--persist)))
 
-(defun org-node-context--persist-file ()
+(defun org-node-context--persistence-file ()
   "Return path to file that caches previews between sessions."
   (mkdir org-node-data-dir t)
-  (file-name-concat org-node-data-dir "org-node-backlink-previews.eld"))
+  (expand-file-name "org-node-backlink-previews.eld" org-node-data-dir))
 
 (defun org-node-context--persist ()
-  "Sync all cached previews to disk."
-  (if org-node-context-persist-on-disk
-      ;; Only proceed if table has gained new entries.
-      ;; NOTE: Does not proceed if there are merely new previews in existing
-      ;;       entries, but it's good enough this way.
-      (when (not (eq org-node-context--last-tbl-state
-                     (hash-table-count org-node-context--previews)))
-        (org-node-context--clean-stale-previews)
-        (setq org-node-context--last-tbl-state
-              (hash-table-count org-node-context--previews))
-        (with-temp-file (org-node-context--persist-file)
-          (let ((print-length nil))
-            (prin1 org-node-context--previews (current-buffer)))))
-    (cancel-timer org-node-context--persist-timer)
-    (setq org-node-context--did-init-persist nil)))
+  "Sync all cached previews with those on disk."
+  (org-node-context--clean-stale-previews)
+  (when (file-readable-p (org-node-context--persistence-file))
+    (with-temp-file (org-node-context--persistence-file)
+      (insert-file-contents (org-node-context--persistence-file))
+      (let ((data (read (current-buffer))))
+        (when (hash-table-p data)
+          (setq org-node-context--previews
+                (map-merge 'hash-table data org-node-context--previews))
+          (org-node-context--clean-stale-previews)))
+      (erase-buffer)
+      (let ((print-length nil))
+        (prin1 org-node-context--previews (current-buffer)))))
+  nil)
 
 (provide 'org-node-context)
 
